% SimpleFit: Fit simple linear+polynomial model.
%
% SimpleFit(x,y,bn)
%    x:   3xN training RAW values.
%    y:   3xN training JPEG values.
%    bn:  Basename to store training data as.
%
% -- Ayan Chakrabarti <ayanc@eecs.harvard.edu>
function SimpleFit(x0,y0,bn,nfact)

if ~exist('nfact')
  nfact = 1;
end;

% Polynomial Degree to use
deg = 7;

% Attach weights to training samples that effectively makes their
% distribution uniform in RAW space.
sgm = size(x0,2).^(-1/3);  sgm = sgm/2/nfact;
uwts = ones(1,size(x0,2));
for i = 1:size(x0,2)
    xk = repmat(x0(:,i),[1 length([i+1:size(x0,2)])]);
    xk = sum( (xk - x0(:,i+1:end)).^2, 1);
    xk = exp(-xk/2/sgm^2);
    uwts(i) = uwts(i)+sum(xk);
    uwts(i+1:end) = uwts(i+1:end) + xk;
end;
uwts = uwts / sum(uwts) * length(uwts); uwts = 1 ./ uwts;
%%%

% Attach weights to samples that downweights JPEG values that are
% too saturated.
sat = max(y0,[],1) - min(y0,[],1); 
sat = max(sat-50,0); wts = exp(-4/50*sat);
wts = max(10^-2,wts);
%%%

% Combine the two weights above
wts = wts .* uwts;

% Initialize nominal value for T, set up training sets.
%    Tmean--- Estimate of linear transform around which to make random
%    steps to find starting points for the iterative optimization.
%    xs{c},ys{c},indwts{c}--- Training data for channel c.
%    T--- Our estimate of the linear 3x3 matrix
%    coeffs--- 3x(deg+1) matrix of polynomial coefficients.
%              (All rows are identical, since we currently assume the same
%               polynomial across channels).

indwts = cell(3,1); T = diag([1 1 1]); coeffs = zeros(3,deg+1);
xs = cell(3,1); ycs = cell(3,1);
for c = 1:3
    % Select non-clipped values for fits
    idx = find(y0(c,:) > 0.5 & y0(c,:) < 254.5);
    xs{c} = x0(:,idx); 
    ycs{c} = y0(c,idx);
    indwts{c} = wts(idx);
    
    % Initialize T
    idx = find(ycs{c} < 255/4 & ycs{c} > 0.5 & (indwts{c} > 0.99)); 
    if length(idx) < 6
      idx = find(ycs{c} < 255/4 & ycs{c} > 0.5);
      tmp = indwts{c}(idx);
      [tmp,idx2] = sort(-tmp);
      idx = idx(idx2(1:6));
    end;
    
    v = pinv((xs{c}(:,idx))')*(ycs{c}(idx)');
    T(c,:) = v';
end;
Tmean = T / sum(T(:))*3;


% Start ranomd restarts
randn('seed',0); % For consistency across runs
BestErr = Inf; randRest = 0;
for random = 1:25

  if random == 1
    T = Tmean;
  else
    T = Tmean + randn(3,3)*0.1; T = T / sum(T(:)) * 3;
  end;
  
  for iters = 1:50
    fprintf('\r%d -- Basic Fitting: Iteration %03d      ',random,iters);
    
    % Do polynomial fit based on T
    xt = []; yc = []; wts = [];
    for c = 1:3
      xt = [xt polyaug(T(c,:)*xs{c},deg)];
      yc = [yc ycs{c}];
      wts = [wts indwts{c}];
    end;
    p = polyfit(xt,yc,deg,wts)';
    
    if sum(1-isfinite(p(:))) > 0
      throw(MException('ForwardFit:NODATA','NODATA'));
    end;
    
    coeffs = repmat(p,[3 1]);

    % Do T fit based on polynomial
    Tnew = T;
    for c = 1:3
      Tnew(c,:) = lincfit(xs{c},ycs{c},p,T(c,:),indwts{c});
    end;
    Tnew = Tnew / sum(Tnew(:)) * 3;
    
    if max(abs(Tnew(:)-T(:))) < 10^-6
      break;
    end;
    T = Tnew;
  end;

  % Compute fit error
  fiterr = 0; den = 0;
  for c = 1:3
    yp = coeffs(c,:)*polyaug(T(c,:)*xs{c},deg);
    fiterr = fiterr + sum(indwts{c} .* (yp - ycs{c}).^2);
    den = den + sum(indwts{c});
  end;
  fprintf('Fit Error = %f\n',sqrt(fiterr/den));
  
  % If better than our best estimate so far, save these.
  if fiterr < BestErr
    save('-mat',[bn '_T.mat'],'T','coeffs','deg');
    fprintf('**** Saving \n');
    BestErr = fiterr;
    Tmean = T;
    randRest = 0;
  else
    % Break if no better result since 10 random restarts
    % since the last time we found a better estimate.
    randRest = randRest + 1;
    if randRest > 10
      break;
    end;
  end;
end;


% Fit the linear part of the model per channel
% x,yc: Find v such that f_p(v*x) = yc
% vinit: Current estimate of v
% wts: Weights of training data points.
function v = lincfit(x,yc,p,vinit,wts)

deg = length(p)-1;

v = vinit; xt = polyaug(v*x,deg); yp = p*xt;

for initers = 1:100
  grad = zeros(1,3);
  for c2 = 1:3
    grad(c2) = sum(wts.* (yp-yc) .* x(c2,:) .* sum(repmat([1:deg]'.*p(2:end)',[1 length(yp)]) .* xt(2:end,:),1));
  end;
  ngrad = -grad / norm(grad);
  if norm(grad) == 0
      break;
  end;
  % Find step-size
  szs = linspace(0,0.5,100);
  scrs = zeros(1,100);
  for j = 1:length(scrs)
    scrs(j) = sqrt(mean( wts.* ((p*polyaug((v+szs(j)*ngrad)*x,deg)-yc).^2) ));
  end;
  sz = szs(find(scrs == min(scrs))); 
  sz = sz(1);
  if sz == 0
    break;
  end;

  % Update v
  v = v+sz*ngrad;
  xt = polyaug(v*x,deg); yp = p*xt;
end;


% Fit polynomial coefficients
% find p such that f_p(xt) = yc
function p = polyfit(xt,yc,deg,wts)

wts = sqrt(wts);
xt = xt .* repmat(wts,[size(xt,1) 1]);
yc = yc .* wts;

options = optimset('LargeScale','off','MaxIter',10000,'Display','off');
xall = polyaug(linspace(0,1,100),deg-1); xall = [xall(1,:); xall];
xall = -xall.*repmat([0:deg]',[1 100]);
xall = xall'; zrs = zeros(100,1);

p = quadprog(xt*xt',-xt*yc',xall,zrs,[],[],[],[],[],options);
    
